import io
from pathlib import Path
import shutil

import boto3
from botocore.utils import calculate_tree_hash

from app import NAME
from app.util import File, JobStatus, OffsetRangeWrapper


PART_SIZE = 64 * 1024 * 1024
MULTIPART_LIMIT = 100 * 1024 * 1024


class Backend():
    """Backend for Amazon Glacier-backed boxes."""

    def __init__(self, box_path, box_config):
        self.box_path = box_path
        self.box_config = box_config
        self.tier = box_config['tier']
        profile = box_config['profile']
        vault = box_config['vault']
        self.session = boto3.session.Session(profile_name=profile)
        self.glacier = self.session.resource('glacier')
        self.vault = self.glacier.Vault('-', vault)

    def box_init(self):
        """Optional box initialization at creation time."""
        self.vault.last_inventory_date  # Quick access test

    def store(self, src_path, name):
        """Store the given file under the name, return a retrieval key."""
        size = src_path.stat().st_size
        if size < MULTIPART_LIMIT:
            with open(src_path, 'rb') as f:
                result = self.vault.upload_archive(
                    archiveDescription=name, body=f)
            archive_id = result.id
        else:
            mpu = self.vault.initiate_multipart_upload(
                archiveDescription=name, partSize=str(PART_SIZE))
            offset = 0
            with open(src_path, 'rb') as f:
                treehash = calculate_tree_hash(f)
                while offset < size:
                    max_offset = min(size, offset + PART_SIZE)
                    r = 'bytes {}-{}/*'.format(offset, max_offset - 1)
                    w = OffsetRangeWrapper(f, offset, max_offset)
                    mpu.upload_part(range=r, body=w)
                    offset += PART_SIZE
            result = mpu.complete(archiveSize=str(size), checksum=treehash)
            archive_id = result['archiveId']
        return archive_id

    def retrieve_init(self, retrieval_key, options):
        """Initiate a retrieval job, return the job key."""
        archive = self.vault.Archive(retrieval_key)
        params = {
            'Type': 'archive-retrieval',
            'ArchiveId': retrieval_key,
            'Tier': self.tier,
        }
        if 'Tier' in options:
            params['Tier'] = options['Tier']
        job = archive.initiate_archive_retrieval(jobParameters=params)
        return job.id

    def retrieve_status(self, *job_keys):
        """Return the overall JobStatus of the given jobs.

           Returns failure if any state is failure,
           success if all states are success,
           running otherwise.
           """
        states = [self._job_status(k) for k in job_keys]
        if any(s == JobStatus.failure for s in states):
            return JobStatus.failure
        elif all(s == JobStatus.success for s in states):
            return JobStatus.success
        else:
            return JobStatus.running

    def retrieve_finish(self, job_key):
        """Finish the job, return the temporary file's Path."""
        job = self.vault.Job(job_key)
        job.load()
        if job.status_code != 'Succeeded':
            raise Exception('Job was not successful.')

        response = job.get_output()
        src = response['body']
        tmp_path = File.mktemp()
        with open(tmp_path, 'wb') as dst:
            shutil.copyfileobj(src, dst, 65536)
        return tmp_path

    def _job_status(self, job_key):
        """Return the JobStatus of the given job."""
        job = self.vault.Job(job_key)
        job.load()
        if job.status_code == 'Succeeded':
            return JobStatus.success
        elif job.status_code == 'Failed':
            return JobStatus.failure
        else:
            return JobStatus.running



import io
import shutil

import gnupg
import yaml

from app.util import File


class GPG():
    """GnuPG support class."""

    def __init__(self, gpg_path=None):
        if gpg_path is None or not gpg_path.is_dir():
            self.gpg = gnupg.GPG(use_agent=True)
        else:
            self.gpg = gnupg.GPG(use_agent=True, gnupghome=str(gpg_path))

    def valid_key_id(self, key_id):
        """Return True if a single keypair for the given ID exists."""
        keys = self.gpg.list_keys(True, keys=key_id)
        return len(keys) == 1

    def encrypt(self, src_path, key_id):
        """Encrypt the source as data and metadata files.

           Return a Path tuple of data and metadata file.
           """
        src_name = src_path.name
        if src_path.is_file():
            src_type = 'file'
        elif src_path.is_dir():
            src_type = 'directory/tar'
        else:
            raise Exception('Unsupported source type.')

        if src_type != 'file':
            src_path = File.tar(src_path)

        try:
            data_path = File.mktemp()
            with open(src_path, 'rb') as src:
                self._encrypt_to_file(src, data_path, key_id)
        except Exception as e:
            msg = 'Source encryption failed. ({})'.format(e)
            raise Exception(msg)
        finally:
            if src_type != 'file' and src_path.exists():
                src_path.unlink()

        sha256 = File.sha256(data_path)
        metadata = {
            'name': src_name,
            'type': src_type,
            'sha256': sha256,
        }
        metayaml = yaml.safe_dump(metadata, default_flow_style=False)
        try:
            src = io.BytesIO(metayaml.encode())
            meta_path = File.mktemp()
            self._encrypt_to_file(src, meta_path, key_id)
        except Exception as e:
            msg = 'Metadata encryption failed. ({})'.format(e)
            raise Exception(msg)

        return data_path, meta_path

    def decrypt(self, data_path, meta_path, dst_path):
        """Recreate the source from encrypted data and metadata."""
        try:
            with open(meta_path, 'rb') as src:
                metayaml = self._decrypt_to_string(src)
        except Exception as e:
            msg = 'Metadata decryption failed. ({})'.format(e)
            raise Exception(msg)

        metadata = yaml.safe_load(metayaml)
        source_name = metadata['name']
        source_type = metadata['type']
        old_sha256 = metadata['sha256']
        new_sha256 = File.sha256(data_path)
        if old_sha256 != new_sha256:
            raise Exception('Source checksum failed.')

        if source_type == 'file':
            source_path = dst_path.joinpath(source_name)
        elif source_type == 'directory/tar':
            source_path = File.mktemp()
        else:
            raise Exception('Unsupported type: ' + str(source_type))

        try:
            with open(data_path, 'rb') as src:
                self._decrypt_to_file(src, source_path)
        except Exception as e:
            msg = 'Source decryption failed. ({})'.format(e)
            raise Exception(msg)

        if source_type == 'directory/tar':
            shutil.unpack_archive(source_path, dst_path, 'tar')
            source_path.unlink()

    def _encrypt_to_file(self, src, dst_path, key_id):
        """Encrypt and sign the given source."""
        key_fp = self._fingerprint(key_id)
        result = self.gpg.encrypt_file(
            src, key_fp, sign=key_fp, armor=False, output=str(dst_path))
        if not result.ok:
            raise Exception(result.status)

    def _fingerprint(self, key_id):
        """Return the fingerprint for the given key ID."""
        keys = self.gpg.list_keys(True, keys=key_id)
        if len(keys) == 0:
            raise Exception('Key not found: ' + str(key_id))
        if len(keys) > 1:
            raise Exception('Ambiguous key ID: ' + str(key_id))
        return keys[0]['fingerprint']

    def _decrypt_to_file(self, src, dst_path):
        """Decrypt and verify the given source, as a file."""
        result = self.gpg.decrypt_file(src, output=str(dst_path))
        self._verify_decrypt_result(result)

    def _decrypt_to_string(self, src):
        """Decrypt and verify the given source, as a string."""
        result = self.gpg.decrypt_file(src)
        self._verify_decrypt_result(result)
        return str(result)

    @staticmethod
    def _verify_decrypt_result(result):
        """Check for successful decryption and a trusted signature."""
        if not result.ok:
            raise Exception(result.status)
        if result.trust_level is None:
            raise Exception('Unsigned data.')
        if result.trust_level < result.TRUST_FULLY:
            raise Exception('Untrusted signature.')



from pathlib import Path

import click

from app import NAME
from app.box import get_box
from app.gpg import GPG


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    '--base-dir', '-b',
    help='Base directory for configuration data.',
    type=click.Path(file_okay=False))
@click.pass_context
def icebox(ctx, base_dir):
    """Store GnuPG-encrypted files in Amazon Glacier."""
    ctx.obj = {}
    if base_dir is None:
        ctx.obj['base'] = Path(click.get_app_dir(NAME))
    else:
        ctx.obj['base'] = Path(base_dir)


@icebox.group()
@click.argument('box-name')
@click.argument('key-id')
@click.pass_context
def init(ctx, box_name, key_id):
    """Create a new box."""
    base_path = ctx.obj['base']
    box = get_box(base_path, box_name)
    if box.exists():
        click.echo('Box already exists.')
        ctx.exit(1)

    gpg = GPG(base_path.joinpath('GPG'))
    if not gpg.valid_key_id(key_id):
        click.echo('Invalid key ID.')
        ctx.exit(1)

    box.key = key_id
    ctx.obj['box'] = box


@init.command('folder')
@click.argument(
    'folder-path',
    type=click.Path(
        exists=True, file_okay=False, resolve_path=True, writable=True))
@click.pass_context
def init_folder(ctx, folder_path):
    """Create a folder-backed box."""
    box = ctx.obj['box']
    box.config['backend'] = 'folder'
    box.config['folder-path'] = folder_path
    try:
        box.init()
    except Exception as e:
        msg = 'Box initialization failed. ({})'.format(e)
        click.echo(msg)
        ctx.exit(1)
    click.echo('Box initialized.')


@init.command('glacier')
@click.argument('vault')
@click.option('--profile', '-p', help='AWS profile', default='default')
@click.option(
    '--tier', '-t',
    help='Default retrieval tier',
    type=click.Choice(['Expedited', 'Standard', 'Bulk']),
    default='Standard')
@click.pass_context
def init_glacier(ctx, vault, profile, tier):
    """Create an Amazon Glacier-backed box."""
    box = ctx.obj['box']
    box.config['backend'] = 'glacier'
    box.config['vault'] = vault
    box.config['profile'] = profile
    box.config['tier'] = tier
    try:
        box.init()
    except Exception as e:
        msg = 'Box initialization failed. ({})'.format(e)
        click.echo(msg)
        ctx.exit(1)
    click.echo('Box initialized.')


@icebox.command()
@click.argument('box-name')
@click.argument('source', type=click.Path(exists=True))
@click.pass_context
def put(ctx, box_name, source):
    """Store data in a box."""
    base_path = ctx.obj['base']
    box = get_box(base_path, box_name)
    if not box.exists():
        click.echo('Box not found.')
        ctx.exit(1)

    src_path = Path(source)
    src_name = src_path.name
    if box.contains(src_name):
        click.echo('Source name already exists in box.')
        ctx.exit(1)

    gpg = GPG(base_path.joinpath('GPG'))
    data_path = None
    meta_path = None
    try:
        data_path, meta_path = gpg.encrypt(src_path, box.key)
        box.store(data_path, meta_path, src_name)
    except Exception as e:
        click.echo(str(e))
        ctx.exit(1)
    finally:
        if data_path and data_path.exists():
            data_path.unlink()
        if meta_path and meta_path.exists():
            meta_path.unlink()
    click.echo('Source stored in box.')


@icebox.command()
@click.argument('box-name')
@click.argument('source')
@click.argument(
    'destination',
    type=click.Path(
        exists=True, file_okay=False, resolve_path=True, writable=True))
@click.option(
    '--option', '-o',
    help='A key=value option for the backend operation.',
    multiple=True)
@click.pass_context
def get(ctx, box_name, source, destination, option):
    """Retrieve data from a box."""
    base_path = ctx.obj['base']
    box = get_box(base_path, box_name)
    if not box.exists():
        click.echo('Box not found.')
        ctx.exit(1)

    if not box.contains(source):
        click.echo('Source name not found in box.')
        ctx.exit(1)

    backend_options = dict(o.split('=') for o in option)
    gpg = GPG(base_path.joinpath('GPG'))
    data_path = None
    meta_path = None
    try:
        data_path, meta_path = box.retrieve(source, backend_options)
        gpg.decrypt(data_path, meta_path, Path(destination))
    except Exception as e:
        click.echo(str(e))
        ctx.exit(1)
    finally:
        if data_path and data_path.exists():
            data_path.unlink()
        if meta_path and meta_path.exists():
            meta_path.unlink()
    click.echo('Source retrieved from box.')



from pathlib import Path
import shutil

from app import NAME
from app.util import File, JobStatus


class Backend():
    """Backend for folder-backed boxes."""

    def __init__(self, box_path, box_config):
        self.box_path = box_path
        self.box_config = box_config
        self.folder_path = Path(box_config['folder-path'])

    def box_init(self):
        """Optional box initialization at creation time."""
        pass

    def store(self, src_path, name):
        """Store the given file under the name, return a retrieval key."""
        dst_path = self.folder_path.joinpath(name)
        shutil.copy(str(src_path), str(dst_path))
        return name

    def retrieve_init(self, retrieval_key, options):
        """Initiate a retrieval job, return the job key."""
        return retrieval_key

    def retrieve_status(self, *job_keys):
        """Return the overall JobStatus of the given jobs.

           Returns failure if any state is failure,
           success if all states are success,
           running otherwise.
           """
        states = [self._job_status(k) for k in job_keys]
        if any(s == JobStatus.failure for s in states):
            return JobStatus.failure
        elif all(s == JobStatus.success for s in states):
            return JobStatus.success
        else:
            return JobStatus.running

    def retrieve_finish(self, job_key):
        """Finish the job, return the temporary file's Path."""
        file_path = self.folder_path.joinpath(job_key)
        tmp_path = File.mktemp()
        shutil.copy(str(file_path), str(tmp_path))
        return tmp_path

    def _job_status(self, job_key):
        """Return the JobStatus of the given job."""
        file_path = self.folder_path.joinpath(job_key)
        if file_path.exists():
            return JobStatus.success
        else:
            return JobStatus.failure



from enum import Enum
import hashlib
import io
import os
from pathlib import Path
import shutil
import sys
import tempfile

from app import NAME


class File():
    """Various file operations."""

    @staticmethod
    def mktemp():
        """Create a temporary file, return its Path."""
        fd, tmpfile = tempfile.mkstemp(prefix=NAME)
        os.close(fd)
        return Path(tmpfile)

    @staticmethod
    def sha256(path):
        """Return the SHA-256 digest of the given file."""
        d = hashlib.sha256()
        with open(path, 'rb') as src:
            while True:
                chunk = src.read(65536)
                if not chunk:
                    break
                d.update(chunk)
        return d.hexdigest()

    @staticmethod
    def tar(path):
        """Create a tar-file from the given Path, return its Path."""
        try:
            tmpfile = File.mktemp()
            tmpfile.unlink()
            base = str(tmpfile)
            tar = shutil.make_archive(base, 'tar', path.parent, path.name)
            return Path(tar)
        except Exception as e:
            msg = 'Tar operation failed. ({})'.format(e)
            raise Exception(msg)


class JobStatus(Enum):
    """Status codes for long-running jobs."""

    running = 0
    success = 1
    failure = 2


class OffsetRangeWrapper():
    """Expose only part of a stream for reading.

       Given a start and end offset, allows reading bytes from start offset to
       end offset - 1 from the wrapped stream. The seek and tell methods use a
       virtual offset ranging from 0 (start offset) to maximum (end offset -
       start offset).
       """
    def __init__(self, stream, start_offset, end_offset):
        assert 0 <= start_offset <= end_offset
        self.stream = stream
        self.start_offset = start_offset
        self.end_offset = end_offset
        stream.seek(start_offset)

    def read(self, size=-1):
        """Read bytes from the offset range."""
        if size is None or size < 0:
            size = sys.maxsize
        current = self.stream.tell()
        if current < self.end_offset:
            actual = min(size, self.end_offset - current)
            return self.stream.read(actual)
        else:
            return bytes()

    def seek(self, offset, whence=io.SEEK_SET):
        """Seek to the given virtual offset in the offset range."""
        if whence == io.SEEK_CUR:
            absolute = self.stream.tell() + offset
        elif whence == io.SEEK_END:
            absolute = self.end_offset + offset
        else:
            absolute = self.start_offset + offset
        actual = min(absolute, self.end_offset)
        actual = max(actual, self.start_offset)
        return self.stream.seek(actual)

    def tell(self):
        """Return the virtual offset in the offset range."""
        current = self.stream.tell()
        virtual = current - self.start_offset
        return virtual



import app.backend.folder
import app.backend.glacier


def get_backend(name, box_path, box_config):
    """Return a backend instance for the given name."""
    if name == 'folder':
        return app.backend.folder.Backend(box_path, box_config)
    elif name == 'glacier':
        return app.backend.glacier.Backend(box_path, box_config)
    else:
        raise Exception('Unsupported backend: ' + name)



from os import path
from setuptools import setup, find_packages

from app import NAME, VERSION


here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=NAME,
    version=VERSION,

    python_requires='>=3.5',
    install_requires=[
        'boto3>=1.7.40',
        'click>=6.7',
        'python-gnupg>=0.4.3',
        'pyyaml>=3.12',
    ],

    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'tmp']),
    entry_points={'console_scripts': ['icebox=app.cli:icebox']},

    author='',
    author_email='',
    description='Encrypting command-line client for Amazon Glacier.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='',
    license='GPLv3+',
)



import time
import uuid

import yaml

from app.backend import get_backend
from app.util import JobStatus


DATA_SUFFIX = '.data'
META_SUFFIX = '.meta'


class Box():
    """Box configuration and mappings."""

    def __init__(self, path):
        self.path = path
        self.config_file = self.path.joinpath('config.yml')
        if self.config_file.exists():
            with open(self.config_file, 'r') as f:
                self.config = yaml.safe_load(f)
        else:
            self.config = {}

    def exists(self):
        """Check if a configuration file for this box exists."""
        return self.config_file.exists()

    def init(self):
        """Initialize this box on creation."""
        if self.exists():
            raise Exception('Box already initialized.')

        backend = self.backend()
        backend.box_init()
        self._save_config()

    @property
    def key(self):
        """Get the GnuPG key ID for this box."""
        return self.config['key-id']

    @key.setter
    def key(self, value):
        """Set the GnuPG key ID for this box."""
        self.config['key-id'] = value

    def _save_config(self):
        """Save the configuration to file."""
        self.path.mkdir(mode=0o770, parents=True, exist_ok=True)
        with open(self.config_file, 'w') as f:
            yaml.safe_dump(self.config, f, default_flow_style=False)

    def store(self, data_path, meta_path, source_name):
        """Store encrypted data and metadata in the backend."""
        data_name, meta_name = _encrypted_names()
        backend = self.backend()
        try:
            data_key = backend.store(data_path, data_name)
            meta_key = backend.store(meta_path, meta_name)
            self._set_retrieval_keys(source_name, data_key, meta_key)
        except Exception as e:
            msg = 'Storage operation failed. ({})'.format(e)
            raise Exception(msg)

    def retrieve(self, source, backend_options):
        """Retrieve encrypted data and metadata from the backend.

           Return a Path tuple of encrypted data and metadata.
           """
        backend = self.backend()
        try:
            # Get existing jobs or start new ones
            if self._has_retrieval_jobs(source):
                data_job, meta_job = self._get_retrieval_jobs(source)
            else:
                data_key, meta_key = self._get_retrieval_keys(source)
                data_job = backend.retrieve_init(data_key, backend_options)
                meta_job = backend.retrieve_init(meta_key, backend_options)
                self._set_retrieval_jobs(source, data_job, meta_job)

            # Wait until jobs are done
            status = backend.retrieve_status(data_job, meta_job)
            while status == JobStatus.running:
                # TODO use logger
                print('Waiting for jobs.')
                time.sleep(60)
                status = backend.retrieve_status(data_job, meta_job)
            self._clear_retrieval_jobs(source)
            if status == JobStatus.failure:
                raise Exception('Retrieval job failed.')

            # Download the files
            data_path = backend.retrieve_finish(data_job)
            meta_path = backend.retrieve_finish(meta_job)
            return data_path, meta_path
        except Exception as e:
            msg = 'Retrieval operation failed. ({})'.format(e)
            raise Exception(msg)

    def contains(self, source):
        """Return True if the source name exists in this box."""
        key_path = self.path.joinpath('retrieval-keys')
        key_file = key_path.joinpath(source)
        return key_file.exists()

    def _get_retrieval_keys(self, source):
        """Get retrieval keys for the given source."""
        key_path = self.path.joinpath('retrieval-keys')
        key_file = key_path.joinpath(source)
        if key_file.exists():
            with open(key_file, 'r') as f:
                keydict = yaml.safe_load(f)
            return keydict.get('data-key'), keydict.get('meta-key')
        else:
            return None, None

    def _set_retrieval_keys(self, source, data_key, meta_key):
        """Set retrieval keys for the given source."""
        key_path = self.path.joinpath('retrieval-keys')
        key_path.mkdir(mode=0o770, parents=True, exist_ok=True)
        key_file = key_path.joinpath(source)
        keydict = {
            'data-key': data_key,
            'meta-key': meta_key,
        }
        with open(key_file, 'w') as f:
            yaml.safe_dump(keydict, f, default_flow_style=False)

    def _has_retrieval_jobs(self, source):
        """Return True if retrieval jobs exist for the given source."""
        job_path = self.path.joinpath('retrieval-jobs')
        job_file = job_path.joinpath(source)
        return job_file.exists()

    def _get_retrieval_jobs(self, source):
        """Get retrieval jobs for the given source."""
        job_path = self.path.joinpath('retrieval-jobs')
        job_file = job_path.joinpath(source)
        if job_file.exists():
            with open(job_file, 'r') as f:
                jobdict = yaml.safe_load(f)
            return jobdict.get('data-job'), jobdict.get('meta-job')
        else:
            return None, None

    def _set_retrieval_jobs(self, source, data_job, meta_job):
        """Set retrieval jobs for the given source."""
        job_path = self.path.joinpath('retrieval-jobs')
        job_path.mkdir(mode=0o770, parents=True, exist_ok=True)
        job_file = job_path.joinpath(source)
        jobdict = {
            'data-job': data_job,
            'meta-job': meta_job,
        }
        with open(job_file, 'w') as f:
            yaml.safe_dump(jobdict, f, default_flow_style=False)

    def _clear_retrieval_jobs(self, source):
        """Clear retrieval jobs for the given source."""
        job_path = self.path.joinpath('retrieval-jobs')
        job_file = job_path.joinpath(source)
        if job_file.exists():
            job_file.unlink()

    def backend(self):
        """Return a backend instance for this box."""
        return get_backend(self.config['backend'], self.path, self.config)


def get_box(base_path, box_name):
    """Return a Box instance for the given base directory and name."""
    box_path = base_path.joinpath(box_name)
    box = Box(box_path)
    return box


def _encrypted_names():
    """Return a tuple of names for encrypted data and metadata files."""
    basename = str(uuid.uuid4())
    data_name = basename + DATA_SUFFIX
    meta_name = basename + META_SUFFIX
    return data_name, meta_name
